#include "common.h"

int debug_level = 0;

void put_timeval( struct timeval *tv, char *res ) {
   sprintf( res, "%02lu:%06lu", ( tv->tv_sec % 60 ), tv->tv_usec );
}

void debug_msg( char *title ) {
   char msg[ 80 ] = "";
   struct timeval tv;
   if( debug_level < 1 ) return;
   gettimeofday( &tv, NULL );
   put_timeval( &tv, msg );
   sprintf( msg + strlen( msg ), " : %s\t: pid=%lu, tid=%lu",
            title, (long)getpid(), pthread_self() );
   printf( "%s\n", msg );
}

//---------------------------------------------------------------------------------

static void one_cycle( void ) {                  // единичный цикл активной работы
   int i, n = 0;
   for( i = 0; i < 5000; i++ )
      n = ( n + 37 ) % 31;
}

static unsigned long calibrate( int milisec ) {  // число вызовов one_cycle() за время milisec
   unsigned long res = 0;
   struct timeval tb, tf;
   gettimeofday( &tb, NULL );
   while( 1 ) {
      one_cycle();
      res++;
      gettimeofday( &tf, NULL );
      long interv = ( tf.tv_sec - tb.tv_sec ) * 1000L +
                    ( tf.tv_usec - tb.tv_usec ) / 1000L;
      if( interv >= milisec ) break;
   }
   return res;
}

double calibrate_for( void ) {                   // калибровка числа вызовов one_cycle() за 1 msec.
   int i, ms = 1;
   double dprev = 0., dcurr = 0.;
   const double eps = 0.02;
   for( i = 0; i < 5; i++ ) {
      unsigned long rep = calibrate( ms );
      dcurr = ( (double)rep ) / ms;
      if( i > 0 ) {
         if( fabs( ( round( dcurr ) - dcurr ) / dcurr ) < eps &&
             fabs( ( dcurr - dprev ) / dcurr ) < eps ) break;
      }
      ms *= 10;
      dprev = dcurr;
   }
   return round( dcurr * 100. ) / 100.;
}

static long repit = 0;                    // число активно повторяемых циклов
static struct timespec msec = { 0, 0 };   // пассивная задержка в msec.

void set_active( unsigned int activ_persent, double calibr ) {
   msec.tv_sec = 0;
   msec.tv_nsec = (long)( 100 - activ_persent ) * 1000000L;
   if( activ_persent != 0 )
      repit = (long)round( activ_persent * calibr );
   if( debug_level > 0) 
      printf( "activ persent=%d repit on slice=%lu delay in slice=%lu\n",
              activ_persent, repit, msec.tv_nsec / 1000000L );
}

void slice( void ) {
   if( msec.tv_sec != 0 && msec.tv_nsec != 0 ) {  // пассивная часть
      struct timespec pause = msec;
      nanosleep( &pause, NULL );
   }
   if( repit != 0 ) {                             // активная часть
      long i;
      for( i = 0; i < repit; i++ ) one_cycle();
   }
   return;
}

//---------------------------------------------------------------------------------
