#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <iostream>
#include <iomanip>

using std::cin;
using std::cout;
using std::endl;
using std::flush;

class chld {
  int fi[ 2 ], // pipe – для ввода в дочернем процессе
      fo[ 2 ]; // pipe – для вывода в дочернем процессе
  pid_t pid;
  char** create( const char *s );
public:
  chld( const char*, int* fdi, int* fdo );
};
// это внутренняя private функция-член : построение списка параметров запуска процесса
char** chld::create( const char *s ) {
  char *p = (char*)s, *f;
  int n;
  for( n = 0; ; n++ ) {
    if( ( p = strpbrk( p, " " ) ) == NULL ) break;
    while( *p == ' ' ) p++;
  };
  char **pv = new char* [ n + 2 ];
  for( int i = 0; i < n + 2; i++ ) pv[ i ] = NULL;
  p = (char*)s;
  f = strpbrk( p, " " );
  for( n = 0; ; n++ ) {
    int k = ( f - p );
    pv[ n ] = new char[ k + 1 ];
    strncpy( pv[ n ], p, k );
    pv[ n ][ k ] = '\0';
    p = f;
    while( *p == ' ' ) p++;
    if( ( f = strpbrk( p, " " ) ) == NULL ) {
      pv[ n + 1 ] = strdup( p );
      break;
    }
  }
  return pv;
};
// вот главное "действо" класса – конструктор, здесь переназначаются
// потоки ввода вывода (SYSIN & SYSOUT), клонируется вызывающий процесс,
// и в нём вызывается новый процесс-клиент со своими параметрами:
chld::chld( const char* pr,  int* fdi, int* fdo ) {
  if( pipe( fi ) || pipe( fo ) ) perror( "pipe" ), exit( EXIT_FAILURE );
  // здесь создаётся список параметров запуска
  char **pv = create( pr );
  pid = fork();
  switch( pid ) {
    case -1: perror( "fork" ), exit( EXIT_FAILURE );
    case  0:   // дочерний клон
      close( fi[ 1 ] ), close( fo[ 0 ] );
      if( dup2( fi[ 0 ], STDIN_FILENO ) == -1 ||
          dup2( fo[ 1 ], STDOUT_FILENO ) == -1 )
         perror( "dup2" ), exit( EXIT_FAILURE );
      close( fi[ 0 ] ), close( fo[ 1 ] );
      // запуск консольного клиента
      if( -1 == execvp( pv[ 0 ], pv ) )
         perror( "execvp" ), exit( EXIT_FAILURE );
      break;
    default:   // родительский процесс
      for( int i = 0;; i++ )
        if( pv[ i ] != NULL ) delete pv[ i ]; else break;
      delete [] pv;
      close( fi[ 0 ] ), close( fo[ 1 ] );
      *fdi = fo[ 0 ];
      int cur_flg;
      // чтение из родительского процесса должно быть в режиме O_NONBLOCK
      cur_flg = fcntl( fo[ 0 ], F_GETFL );
      if( -1 == fcntl( fo[ 0 ], F_SETFL, cur_flg | O_NONBLOCK ) )
         perror( "fcntl" ), exit( EXIT_FAILURE );
      *fdo = fi[ 1 ];
      // для записи O_NONBLOCK не обязательно
      break;
  };
};

static int debug_level = 0;
static u_long buflen = 1024;
static u_char *buf;

int main( int argc, char *argv[] ) {
   int c;
   double stret = 1.0;
   while( -1 != ( c = getopt( argc, argv, "vs:b:h" ) ) )
      switch( c ) {
         case 's':
            if( 0.0 != atof( optarg ) ) stret = atof( optarg );
            break;
         case 'b': if( 0 != atol( optarg ) ) buflen = atol( optarg ); break;
         case 'v': debug_level++; break;
         case 'h':
         default : cout <<
                     argv[ 0 ] << "[<опции>] <имя вх.файла> <имя вых.файла>\n"
                     "опции:\n"
                     " -s - вещественный коэффициент темпо-коррекции\n"
                     " -b - размер аудио буфера\n"
                     " -v - увеличить уровень детализации отладочного вывода\n"
                     " -h — вот этот текст подсказки\n";
            exit( 'h' == c ? EXIT_SUCCESS : EXIT_FAILURE );
      }
   if( optind != argc - 2 )
      cout << "должно быть указаны имена входного и выходного звуковых файлов"
           << endl, exit( EXIT_FAILURE );
   // файл с которого читается входной аудиопоток
   int fai = open( argv[ optind ], O_RDONLY );
   if( -1 == fai ) perror( "open input" ), exit( EXIT_FAILURE );
   // файл в который пишется результирующий аудиопоток
   int fao = open( argv[ optind + 1 ], O_RDWR | O_CREAT,  // 666
                   S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH |  S_IWOTH ); 
   if( -1 == fao ) perror( "open output" ), exit( EXIT_FAILURE );
   char stretch[ 80 ] = "";
   if( 1.0 != stret ) sprintf( stretch, " stretch %f", stret );
   char comstr[ 120 ] = "sox -V -twav - -twav - ";
   strcat( comstr, stretch );
   // сформирована командная строка дочернего процесса
   if( debug_level > 1 ) cout << comstr << endl;
   int fdi, fdo;
   chld *ch = new chld( comstr, &fdi, &fdo );
   // дескриптор с которого читается вывод в stdout дочернего процесса
   if( -1 == fdi ) perror( "pipe output" ), exit( EXIT_FAILURE );
   // дескриптор куда записывается то, что читает из stdin дочерний процесс
   if( -1 == fdo ) perror( "pipe output" ), exit( EXIT_FAILURE );
   buf = new u_char[ buflen ];
   int sum[] = { 0, 0, 0, 0 };
   while( true ) {
      int n;
      if( fai > 0 ) {
         n = read( fai, buf, buflen );
         sum[ 0 ] += n > 0 ? n : 0;
         if( debug_level > 2 ) 
            cout << "READ from audio\t" << n << " -> " << sum[ 0 ] << endl;
         if( -1 == n ) perror( "read file" ), exit( EXIT_FAILURE );
         if( 0 == n ) close( fai ), fai = -1;
      };
      if( fai > 0 ) {
         n = write( fdo, buf, n );
         sum[ 1 ] += n > 0 ? n : 0;
         if( debug_level > 2 )
            cout << "WRITE to child\t" << n << " -> "
                 << ( sum[ 1 ] += n > 0 ? n : 0 ) << endl;
         if( -1 == n ) perror( "write pipe" ), exit( EXIT_FAILURE );
         // передеспетчеризация - дать время на обработку
         usleep( 100 );
      }
      else close( fdo ), fdo = -1;
      n = read( fdi, buf, buflen );
      if( debug_level > 2 )
         cout << "READ from child\t" << n << " -> " 
              << ( sum[ 2 ] += n > 0 ? n : 0 ) << flush;
      if( n >= 0 && debug_level > 2 ) cout << endl;
      // это может быть только после закрытия fdo!!!
      if( 0 == n ) break;
      else if( -1 == n ) {
         if( EAGAIN == errno ) {
            if( debug_level > 2 ) 
               cout << " : == not ready == ... wait ..." << endl;
            usleep( 300 );
            continue;
         }
         else perror( "\nread pipe" ), exit( EXIT_FAILURE );
      }
      n = write( fao, buf, n );
      if( debug_level > 2 )
         cout << "WRITE to file\t" << n << " -> " 
              << ( sum[ 3 ] += n > 0 ? n : 0 ) << endl;
      if( -1 == n ) perror( "write file" ), exit( EXIT_FAILURE );
   };
   close( fai ), close( fao );
   close( fdi ), close( fdo );
   delete [] buf;
   delete ch;
   cout << "считано со входа " << sum[ 0 ] << " байт - записано на выход " 
        << sum[ 0 ] << " байт" << endl;
   return EXIT_SUCCESS;
};

