#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

static int debug_level = 0;
static char stretch[ 80 ] = "";
u_long buflen = 1024;
u_char *buf;

// конвейер, которым мы читаем RAW файлы (PCM 16-бит), пример :
// $ cat male.raw | sox -u -s -b16 -r8000 -traw - -t alsa default stretch 0.9
// конвейер, которым мы читаем WAV файлы (или OGG Vorbis), пример:
// $ sox -V male.wav -traw -u -sw - | sox -u -s -b16 -r8000 -traw - -t alsa default stretch 0.9
// конвейер, которым мы читаем OGG SPEEX файлы, пример:
// $ speexdec -V male.spx - | sox -u -s -b16 -r8000 -traw - -t alsa default stretch 0.9

void play_file( char *filename ) {
   struct stat sbuf;
   if( stat( filename, &sbuf ) < 0 ) {
      fprintf( stdout, "неверное имя файла: %s\n", filename );
      return;
   }
   // форматы файла различаются по имени, но должны бы ещё и по magic
   // содержимому с начала файла: "RIFF..." - для *.wav, "Ogg ... vorbis" - для
   // *.ogg, "Ogg ... Speex" — для *.spx, или отсутствие magic признаков
   // для *.pcm, *.raw
   const char *ftype[] = { ".raw", ".pcm", ".wav", ".ogg", ".spx" };
   int stype = sizeof( ftype ) / sizeof( ftype[ 0 ] ), i;
   for( i = 0; i < stype; i++ ) {
      char *ext = strstr( filename, ftype[ i ] );
      if( NULL == ext ) continue;
      if( strlen( ext ) == strlen( ftype[ i ] ) ) break;
   }
   if( i == stype ) {
      fprintf( stdout, "неизвестный формат аудио файла: %s\n", filename );
      return;
   };
   char cmd[ 120 ];
   const char *inpcmd[] = {
      "cat %s",
      "sox%s %s -traw -u -s -",
      "speexdec%s %s -"
   };
   const int findex[] = { 0, 0, 1, 1, 2 };
   const char* cmdfmt = inpcmd[ findex[ i ] ];
   if( 0 == findex[ i ] )
      sprintf( cmd, cmdfmt, filename );
   else if( 1 == findex[ i ] )
      sprintf( cmd, cmdfmt,
               0 == debug_level ? " -q" : debug_level > 1 ? " -V" : "",
               filename, stretch );
   else
      sprintf( cmd, cmdfmt, debug_level > 1 ? " -V" : "", filename );
   if( debug_level > 1 ) fprintf( stdout, "%s\n", cmd );
   FILE *fsox = popen( cmd, "r" );
   const char *outcmd = "sox%s -u -s -b16 -r8000 -traw - -t alsa default %s";
   sprintf( cmd, cmdfmt = outcmd,
            0 == debug_level ? " -q" : debug_level > 1 ? " -V" : "",
            stretch );
   if( debug_level > 1 ) fprintf( stdout, "%s\n", cmd );
   FILE *fplay = popen( cmd, "w" );
   int in, on, s = 0;
   while( ( in = fread( buf, 1, buflen, fsox ) ) != 0 ) {
      if( debug_level ) fprintf( stdout, "read : %d - ", in ), fflush( stdout );
      on = fwrite( buf, 1, in, fplay );
      if( debug_level ) fprintf( stdout, "write : %d\n", on ), fflush( stdout );
      s += on;
   }
   if( debug_level ) fprintf( stdout, "воспроизведено: %d байт\n", s );
}

int main( int argc, char *argv[] ) {
   int c;
   double stret = 1.0;
   while( -1 != ( c = getopt( argc, argv, "vs:b:" ) ) )
      switch( c ) {
         case 's':
            if( 0.0 != atof( optarg ) ) stret = atof( optarg );
            break;
         case 'b': if( 0 != atol( optarg ) ) buflen = atol( optarg ); break;
         case 'v': debug_level++; break;
         case 'h':
         default :
            fprintf( stdout,
                     "Опции:\n"
                     " -s - вещественный коэффициент темпо-коррекции\n"
                     " -b - размер аудио буфера\n"
                     " -v - увеличить уровень детализации отладочного вывода\n"
                     " -h - вот этот текст подсказки\n" );
            exit( 'h' == c ? EXIT_SUCCESS : EXIT_FAILURE );
      }
   if( optind == argc )
      fprintf( stdout, "должен быть указан хотя бы один звуковой файл\n" ),
      exit( EXIT_FAILURE );
   if( 1.0 != stret ) sprintf( stretch, " stretch %f", stret );
   buf = malloc( buflen );
   int i;
   for( i = optind; i < argc; i++ ) play_file( argv[ i ] );
   free( buf );
   return EXIT_SUCCESS;
};

