#define _GNU_SOURCE
#include "common.h"
#include <sys/wait.h>

char sout[ 1000 ], *pout = &sout[ 0 ];
pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
char base;

#define NREP 20   // число циклов выполнения каждого потока
void* threadfunc ( void* data ) {
   int id = (long)data, i;
   for( i = 0; i < NREP; i++ ) {
      delay( 5 ); // пассивная пауза 5 миллисек.
      pthread_mutex_lock( &lock );
      *pout++ = base + id;
      pthread_mutex_unlock( &lock );
   }
   pthread_exit( NULL );
   return NULL;
};

void prepare( void ) { pthread_mutex_lock( &lock ); }

void parent( void ) { pthread_mutex_unlock( &lock ); }

static int npth;  // число дочерних потоков
pthread_t *tid;   // массив ID потоков

void child( void ) { 
   int i;
   pthread_attr_t attr;
   pthread_attr_init( &attr );
   for( i = 0; i < npth; i++ ) { // перезапуск потоков в дочернем процессе
      pthread_getattr_np( tid[ i ], &attr ); 
      pthread_create( tid + i, &attr, threadfunc, (void*)(long)i );
   }
   pthread_attr_destroy( &attr );
   pthread_mutex_unlock( &lock );
}

int main( int argc, char *argv[] ) {
   int i;
   npth = ( argc > 1 && atoi( argv[ 1 ] ) > 0 ) ?
              atoi( argv[ 1 ] ) : 3;
   tid = (pthread_t*)calloc( npth, sizeof( pthread_t ) );
   for( i = 0; i < npth; i++ )   // запуск потоков в родительском процессе
      pthread_create( tid + i, NULL, threadfunc, (void*)(long)i );
   pthread_atfork( prepare, parent, child );
   pid_t pid = fork();
   if( pid < 0 ) perror( "fork error" ), exit( EXIT_FAILURE );
   else base = pid > 0 ? '0' : 'a';
   for( i = 0; i < npth; i++ )   // ожидание завершения всех
      pthread_join( tid[ i ], NULL );
   *pout = '\0';
   printf( "%s\n", sout );
   free( tid );
   if( pid != 0 ) wait( NULL );
   exit( EXIT_SUCCESS );
};
