#include "common.h"

// демонстрация возникновения и изучение инверсии приоритетов: 

void one_cycle( void ) {     // функция активной работы
   int i, n = 0;
   for( i = 0; i < 50; i++ ) // цикл произвольной активной работы
      n = ( n + 37 ) % 31;
}

/* struct timeval {
      time_t      tv_sec;     // seconds 
      suseconds_t tv_usec;    // microseconds 
   }; */
ulong cycle2msec( void (*func)(void) ) { // калибровка числа вызовов за msec.
   ulong n = 0;
   struct timeval tb, tf;
   gettimeofday( &tb, NULL );            // начальная отметка
   do {
      (*func)();
      gettimeofday( &tf, NULL );         // текущая отметка
      timersub( &tf, &tb, &tf );
      n++;
   } while ( tf.tv_usec <= 1000L );
   return n;
}

ulong c2m; // откалиброванное число вызовов рабочей функции на 1 миллисекунду

void adelay( ulong msec ) { // активная задержка в миллисекундах
   ulong i;
   void (*run)(void) = &one_cycle;  // устанавливается пробная функция
   for( i = 0; i < c2m * msec; i++ ) (*run)();
}

void pdelay( ulong msec ) { // пассивная задержка в миллисекундах
   struct timespec pause = { 0, 0 };
   pause.tv_nsec = msec * 1000000L;
   nanosleep( &pause, NULL );
}

typedef void (*dfunc_t)( ulong );
dfunc_t delay = adelay;     // функция используемая в качестве работы

pthread_mutex_t loc;
struct timeval tb;          // время старта потоков

typedef enum phase { CREATE, START, INWORK, FINAL } phase_t;
typedef struct {
   pthread_t   tid;
   const char *title;
   int         prior;
   void*     (*func)(void*);
   uint        before, after; 
   bool        blocked;
   phase_t     stage;
} defthr_t;

void* funcs( void* );       // функции потока
typedef void* (*tfunc_t)(void*);
tfunc_t lfun = &funcs, mfun = &funcs, hfun = &funcs;

defthr_t threads[] = {
   { 0, "LOW",    1, lfun, 2, 7, true,  CREATE },
   { 0, "MIDLE", 30, mfun, 6, 7, false, CREATE },
   { 0, "HIGH",  60, hfun, 4, 5, true,  CREATE } 
};
const int nthr = sizeof( threads ) / sizeof( threads[ 0 ] ); 

void put_msg( defthr_t* );  // вывод строки диагностики состояния потока

uint multiplier = 1;        // масштаб всех временных интервалов
pthread_barrier_t bstart;

void* funcs( void* par ) {  // функция потоков с возможностью захвата мютекса
   defthr_t rec;                        // копия блока параметров
   memcpy( &rec, (defthr_t*)par, sizeof( defthr_t ) );
   defthr_t *prec = &rec;
   pthread_barrier_wait( &bstart );     // одновременный старт потоков          
   put_msg( prec );
   pdelay( prec->before * multiplier ); // пауза перед началом выполнения
   if( prec->blocked ) pthread_mutex_lock( &loc );
   prec->stage = START;
   put_msg( prec );
   prec->stage = INWORK;
   uint i;
   for( i = 0; i < prec->after * multiplier; i++ ) {   
      delay( multiplier );              // время выполнения шага потока
      put_msg( prec );
   }
   if( prec->blocked ) pthread_mutex_unlock( &loc );
   prec->stage = FINAL;
   put_msg( prec );
   return NULL;
}

int debug_level = 0,                           // уровень диагностического вывода
    mproto = PTHREAD_PRIO_NONE,                // протокол приоритета на мютексе
// either PTHREAD_PRIO_NONE, PTHREAD_PRIO_INHERIT, or PTHREAD_PRIO_PROTECT
    ceiling = 40;                              // граничный приоритет 
static const char *switches = "abnipc:r:m:vh"; // опции запуска
void help( const char *prog );                 // подсказка опций
void mutex_init( void );                       // инициализация мютекса

int main( int argc, char *argv[] ) {
   int opt, ncpu = 1; 
   while ( ( opt = getopt( argc, argv, switches  ) ) != -1 ) {
      switch( opt ) {
         case 'a' : delay = adelay; break;
         case 'b' : delay = pdelay; break;
         case 'n' : mproto = PTHREAD_PRIO_NONE;    break;
         case 'i' : mproto = PTHREAD_PRIO_INHERIT; break;
         case 'p' : mproto = PTHREAD_PRIO_PROTECT; break;
         case 'c' : if( atoi( optarg ) > 0 ) {
                       mproto = PTHREAD_PRIO_PROTECT;
                       ceiling = atoi( optarg );
                       break;
                    }
         case 'r' : if( atoi( optarg ) > 0 ) ncpu = atoi( optarg );
                    break;
         case 'm' : if( atoi( optarg ) > 0 )
                       multiplier = atoi( optarg );
                    break;
         case 'v' : debug_level++; break;
         case 'h' : help( argv[ 0 ] ); break;
         default  : help( argv[ 0 ] ); exit( EXIT_FAILURE );
       }
   };
   struct sched_param param = { .sched_priority = 1 };
   if( sched_setscheduler( 0, SCHED_RR, &param ) != 0 )
      cout << "you must be root!" << endl, exit( EXIT_FAILURE );
   { cpu_set_t set;                // установка числа процессоров [1]
     CPU_ZERO( &set );
     for( int i = 1; i <= ncpu; i++ ) CPU_SET( i, &set);
     if( sched_setaffinity( 0, sizeof( cpu_set_t ), &set ) != 0 )
         perror( "sched_setaffinity" );
     if( debug_level ) {
        if( sched_getaffinity( 0, sizeof( cpu_set_t ), &set ) != 0 )
           perror( "sched_getaffinity" );
        cout << "число используемых процессоров: " << CPU_COUNT( &set ) << endl;
     }
   }
   c2m = cycle2msec( one_cycle );  // калибровка функции активной работы
   if( debug_level > 1 ) {
      cout << "циклов пробной функции в миллисекунде: " << c2m << endl;
      struct timespec tp = { 0, 0 };
      if( 0 == sched_rr_get_interval( 0, &tp ) )
         cout << "интервал SCHED_RR квантования: " << tp.tv_nsec << " nsec." << endl; 
      else cout << "ошибка считывания интервала SCHED_RR квантования" << endl;
   }
   mutex_init(); 
   pthread_barrier_init( &bstart, NULL, nthr + 1 ); // спусковой механизм
   cout << "msec.   :" << endl;
   for( int i = 0; i < nthr; i++ ) {
      defthr_t *p = threads + i; 
      if( pthread_create( (pthread_t*)p, NULL, *( p->func ), (void*)p ) != 0 )
         perror( "pthread_create" );
      param.sched_priority = p->prior;
      if( pthread_setschedparam( p->tid, SCHED_RR, &param ) != 0 ) 
         printf( "pthread_setschedparam" );
      delay( 1 );
   }
   gettimeofday( &tb, NULL );                       // отметка старта потоков
   pthread_barrier_wait( &bstart );                 // одновременный старт потоков          
   for( int i = 0; i < nthr; i++ ) 
      pthread_join( threads[ i ].tid, NULL );
   exit( EXIT_SUCCESS );
};

//---------------- вспомогательные функции -----------------

void put_msg( defthr_t* p ) {
   static pthread_mutex_t mout = PTHREAD_MUTEX_INITIALIZER;
   struct timeval tf;
   gettimeofday( &tf, NULL );    // текущая отметка времени
   timersub( &tf, &tb, &tf );
   long interv = tf.tv_sec * 1000L + tf.tv_usec / 1000L +
                 ( tf.tv_usec % 1000 > 500 ? 1 : 0 ),
        mksec  = ( tf.tv_usec % 1000 ) / 100;

   char title[ 30 ];
   sprintf( title, "%04ld.%02ld : поток %-7s:", interv, mksec, p->title );
   pthread_mutex_lock( &mout );  // разграничение вывода
   switch( p->stage ) {
      case CREATE : printf( "%s ------> create\n", title );
                    break;
      case START  : printf( "%s -------> start work...\n", title );
                    break;
      case FINAL  : printf( "%s -------> finished!\n", title );
                    break;
      case INWORK : {
         int policy; 
         struct sched_param param;
         pthread_getschedparam( p->tid, &policy, &param ); 
         int prioceiling;
         if( pthread_mutex_getprioceiling( &loc, &prioceiling ) != 0 )
            printf( "%s приоритет RT потока = %02d\n",
                    title, param.sched_priority );
         else
            printf( "%s приоритет RT потока = %02d - гран. приоритет = %d\n",
                    title, param.sched_priority, prioceiling );
         break;
      } 
   }
   pthread_mutex_unlock( &mout );
   return;
}

void help( const char* prog ) {
   const char* mean[] = {
      "функция работы: активное ожидание",
      "функция работы: пассивное ожидание",
      "протокол мютекса: невмешательство",
      "протокол мютекса: наследование приоритета",
      "протокол мютекса: граничный приоритет",
      "значение граничного приоритета",
      "число процессоров",
      "мультиплексор шкалы времени, msec.",
      "повысить уровень детализации вывода",
      "подсказка по запуску программы"
   };
   int nopt = strlen( switches ); 
   char* ps = (char*)switches;
   while( 1 ) {
      ps = strchr( ps, ':' );
      if( NULL == ps ) break;
      ps++;
      nopt--;
   }; 
   if( debug_level > 1 ) 
      cout << "число допустимых опций программы = " << nopt << endl;      
   if( sizeof( mean ) / sizeof( mean[ 0 ] ) != nopt )
      cout << "несоответствие опций" << endl, exit( EXIT_FAILURE );
   cout << "Использование: " << prog 
        << " [-a|-b] [-n|-i|-p [-c...]]  [-r...] [-m...] [-v[v]] [-h]" << endl
        << "Ключи:" << endl;
   ps = (char*)switches;
   for( int i = 0; i < nopt; i++ ) {
      cout << "  -" << *ps++;
      if( ':' == *ps ) {
         cout << "<значение>";
         ps++;
      }
      cout << " - " << mean[ i ] << endl;
   }
   exit( EXIT_SUCCESS );
};

void mutex_init( void ) {   // инициализация мютекса под протокол
   pthread_mutexattr_t mattr;
   pthread_mutexattr_init( &mattr );
   if( pthread_mutexattr_setprotocol( &mattr, mproto ) != 0 )
      perror( "pthread_mutexattr_setprotocol" );
   if( pthread_mutex_init( &loc, &mattr ) != 0 )
      perror( "pthread_mutex_init" ), exit( EXIT_FAILURE );
   pthread_mutexattr_destroy( &mattr );
   if( debug_level ) {
      cout << "протокол мютекса: " 
           << ( PTHREAD_PRIO_NONE    == mproto ? "PTHREAD_PRIO_NONE" : 
                PTHREAD_PRIO_INHERIT == mproto ? "PTHREAD_PRIO_INHERIT" : 
                PTHREAD_PRIO_PROTECT == mproto ? "PTHREAD_PRIO_PROTECT" :
                "UNKNOWN" ) << endl;
   } 
   int prioceiling;
   if( ( errno = pthread_mutex_setprioceiling( &loc, ceiling, &prioceiling ) != 0 ) )
      perror( "pthread_mutex_setprioceiling" );
   if( 0 == errno && debug_level > 1 ) {
      cout << "граничный приоритет - было: " << prioceiling;
      pthread_mutex_getprioceiling( &loc, &prioceiling );
      cout << ", стало: " << prioceiling << endl;
   }
}
