#include "common.h"

class dbase : public list<element> {        // защита спин-блокировкой
   static const int READ_DELAY = 1, WRITE_DELAY = 2;
     pthread_spinlock_t loc; 
public: 
   dbase( void ) { pthread_spin_init( &loc, 1 ); }; 
   ~dbase( void ) { pthread_spin_destroy( &loc ); };
   void add( const element& e, bool wait = true ) {
      pthread_spin_lock ( &loc );
      int pos = size() * rand() / RAND_MAX; // вставить в случайную позицию
      list<element>::iterator p = begin();
      for( int i = 0; i < pos; i++ ) p++;
      insert( p, e );
      if( wait ) delay( WRITE_DELAY );
      pthread_spin_unlock ( &loc );
   };
   int pos( const element& e ) {
      uint n = 0;
      pthread_spin_lock ( &loc );
      for( list<element>::iterator i = begin(); i != end(); i++, n++ ) 
         if( *i == e ) { delay( READ_DELAY ); break; };
      if( n == size() ) n = -1;
      pthread_spin_unlock ( &loc );
      return n;
   };
} data;

static void* add( void* par ) {             // функция потока
   data.add( (element)par );
   return NULL;
};

static void* pos( void* par ) {             // функция потока
   data.pos( (element)par );
   return NULL;
};

int main( int argc, char *argv[] ) {
   params( argc, argv );
   cout << "wait ..." << flush;
   for( ulong i = 0; i < n; i++ )
      data.add( erand( n ), false );       // начальное заполнение
   pthread_t *h = new pthread_t[ n ];
   struct timeval tb, tf;
   gettimeofday( &tb, NULL );
   ulong i;
   int ret = 0;
   for( i = 0; i < n; i++ ) {              // параллельная обработка
      element e = erand( n );
      ret = pthread_create( h + i, NULL, wrand( p ) ? &add : &pos, (void*)e );
      if( ret != 0 ) {
         n = i;
         break;
      }
   };
   for( i = 0; i < n; i++ ) pthread_join( h[ i ], NULL );
   gettimeofday( &tf, NULL );
   cout << "\r";
   if( ret != 0 ) 
      cout << "потоков: " << n << ", причина: " << strerror( ret ) << endl;      
   delete [] h;
   timersub( &tf, &tb, &tf );
   long interv = tf.tv_sec * 1000L + tf.tv_usec / 1000L +
                 ( tf.tv_usec % 1000 > 500 ? 1 : 0 );
   cout << "интервал выполнения: " << interv << " миллисекунд" << endl;
   return EXIT_SUCCESS;
};
