#include "common.h"
#include <semaphore.h>
#include <fcntl.h>           /* For O_* constants */

unsigned long N = 1000;   // число циклов
uint T = 2;               // число потоков
static pthread_barrier_t bstart;
static sem_t *sem;
static bool debug = false;
static char *str;
static volatile int ind = 0;

void* threadfunc ( void* data ) {
   pthread_barrier_wait( &bstart );
   unsigned long i = 0;
   char cid = '0' + (long)data;
   uint64_t t = 0, t1; 
   while( i++ != N ) {
      t1 = rdtsc();      
      sem_wait( sem );
      if( debug ) str[ ind++ ] = cid;
      sem_post( sem );
      t += rdtsc() - t1;
      sched_yield();
   };
   sem_wait( sem );
   cout << pthread_self() << ": тактов = " << t << ", на 1 семафор = " << t / N << endl;
   sem_post( sem );
   return NULL;
};

int main( int argc, char *argv[] ) {
   int opt;
   while ( ( opt = getopt( argc, argv, "n:t:v" ) ) != -1 ) {
      switch( opt ) {
         case 'n' :
            if( atol( optarg ) > 0 ) N = atol( optarg );
            break;
         case 't' :
            if( atoi( optarg ) > 0 ) T = atoi( optarg );
            break;
         case 'v' :
            debug = true;
            break;
         default : exit( EXIT_FAILURE );
       }
   };
   if( debug ) str = new char [ T * N + 1 ];
   pthread_t* tid = new pthread_t[ T ];
   if( pthread_barrier_init( &bstart, NULL, T ) != 0 ) perror( "barrier init" ), exit( EXIT_FAILURE );
   const char semname[] = "synchro";
   if( ( sem = sem_open( semname, O_CREAT, S_IRWXO, 1 ) ) == SEM_FAILED )
      perror( "semaphore init" ), exit( EXIT_FAILURE );   
   ulong i;
   for( i = 0; i < T; i++ )
      if( pthread_create( tid + i, NULL, threadfunc, (void*)i ) != 0 )
         perror( "thread create error" ), exit( EXIT_FAILURE );
   for( i = 0; i < T; i++ ) pthread_join( tid[ i ], NULL );
   sem_close( sem ); 
   sem_unlink( semname );   
   if( debug ) {
      str[ ind ] = '\0';
      cout << str << endl;
      delete [] str;
   };
   delete [] tid;
   exit( EXIT_SUCCESS );
};
