#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

typedef struct data_bloc {               // наш собственный тип данных
   pthread_t tid;
   pthread_mutex_t mid;
   int data;   
} data_t;

pthread_key_t key;
data_t global;                           // глобальный экземпляр блока данных

void put_msg( int local, int stat, long param ) {
   printf( "local=%d, global=%d, static=%d, parameter=%ld, own=%lu\n",
           local, global.data, stat,
           param, ((data_t*)pthread_getspecific( key ))->tid );
}

static pthread_once_t once = PTHREAD_ONCE_INIT;

static void destructor( void* db ) {     // деструктор собственных данных
   data_t *p = (data_t*)db;
   free( p );
}

static void once_creator( void ) {       // создаёт единый на процесс ключ для данных data_t
   pthread_key_create( &key, destructor );
}

void* thread_proc( void *parm ) {        // функция потока
   long param = (long)parm;              // переданный параметр
   int local = 0;                        // локальная переменная
   static int data = 0;                  // статическая переменная
   pthread_mutex_lock( &global.mid );
   global.data++; data++;
   pthread_mutex_unlock( &global.mid );
   local++;
   pthread_once( &once, once_creator );  // гарантия единичности создания ключа
   pthread_setspecific( key, malloc( sizeof( data_t ) ) );
   data_t *tsd = pthread_getspecific( key );
   tsd->tid = pthread_self();
   pthread_mutex_lock( &global.mid );
   put_msg( local, data, param );
   pthread_mutex_unlock( &global.mid );
   return NULL;
}

#define TCNT 5                           // число потоков
int main( int argc, char **argv, char **envp ) {
   pthread_t tid[ TCNT ];
   long i;
   global.data = 0;
   pthread_mutex_init( &global.mid, NULL ); 
   for( i = 0; i < TCNT; i++ )
      pthread_create( &tid[ i ], NULL, thread_proc, (void*)( i + 1 ) );
   for( i = 0; i < TCNT; i++ ) 
      pthread_join( tid[ i ], NULL );
   return( EXIT_SUCCESS );
}
