#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <pthread.h>

int quitflag = 0;
sigset_t mask;

pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t wait = PTHREAD_COND_INITIALIZER;

void* threadfunc ( void* data ) {
   int signo;
   while( 1 ) {
      if( sigwait( &mask, &signo ) != 0 )
         perror( "sigwait:" ), exit( EXIT_FAILURE ); 
      switch( signo ) {
         case SIGINT:
            printf( "  ... signal SIGINT\n" );  
            break;
         case SIGQUIT:
            printf( "  ... signal SIGQUIT\n" );  
            pthread_mutex_lock( &lock );
            quitflag = 1;
            pthread_mutex_unlock( &lock );
            pthread_cond_signal( &wait );
            return NULL;
         default:
            printf( "undefined signal %d\n", signo ), exit( EXIT_FAILURE );  
      }
   };
}; 

int main() {
   printf( "process started with PID=%d\n", getpid() );
   sigemptyset( &mask );
   sigaddset( &mask, SIGINT );
   sigaddset( &mask, SIGQUIT );
   sigset_t oldmask;
   if( sigprocmask( SIG_BLOCK, &mask, &oldmask ) < 0 )
      perror( "signals block:" ), exit( EXIT_FAILURE ); 
   pthread_t tid;
   if( pthread_create( &tid, NULL, threadfunc, NULL ) != 0 )
      perror( "thread create:" ), exit( EXIT_FAILURE ); ; 
   pthread_mutex_lock( &lock );
   while( 0 == quitflag )
      pthread_cond_wait( &wait, &lock );
   pthread_mutex_unlock( &lock );
   /* SIGQUIT был перехвачен, но к этому моменту снова заблокирован */
   if( sigprocmask( SIG_SETMASK, &oldmask, NULL ) < 0 )
      perror( "signals set:" ), exit( EXIT_FAILURE ); 
   return EXIT_SUCCESS;
};

